;*****************************************************************************
;*** Name of the program:		TWI (Single) Master Implementation
;*** Function:             	Implementaion of I2C-Bus with connection to a
;***					   			I/O-Extension PCF8574
;*** Author:                	Andreas Hesse
;*** Date of construction:  	2003-04-09
;*** Date of last revision: 	2003-04-11
;*** Version:               	1.0.0.2
;*** Target:						AT90Sxxxx (any AVR device)
;*** Comment: 
;***
;***	Abbreviations:
;***	1. RI = Remote Input
;***	2. RO = Remote Output
;******************************************************************************
;***
;*** Version management:
;*** V 1.0.0.2  Optimized version: Read and write procedure are not separted
;***				 during write procedure for inputs also the output value will
;***				 be written	
;*** V 1.0.0.1  Implement connection to PCF8574
;*** V 1.0.0.0  First construction based on  Version 1.0 of
;*** 		 A P P L I C A T I O N   N O T E   A V R 3 0 0 (2002.05.03)
;******************************************************************************
;*** DESCRIPTION
;*** 	Basic routines for communicating with TWI slave devices. This
;***	"single" master implementation is limited to one bus master on the
;***	TWI bus. Most applications do not need the multimaster ability
;***	the TWI bus provides. A single master implementation uses, by far,
;***	less resources and is less XTAL frequency dependent.
;***
;***	Some features :
;***	* All interrupts are free, and can be used for other activities.
;***	* Supports normal and fast mode.
;***	* Supports both 7-bit and 10-bit addressing.
;***	* Supports the entire AVR microcontroller family.
;***
;***	Main TWI functions :
;***	'TWI_start' -	Issues a start condition and sends address
;***						and transfer direction.
;***	'TWI_rep_start' -	Issues a repeated start condition and sends
;***							address and transfer direction.
;***	'TWI_do_transfer' -	Sends or receives data depending on
;***								direction given in address/dir byte.
;***	'TWI_stop' -	Terminates the data transfer by issue a
;***						stop condition.
;***
;*** USAGE
;***	Transfer formats is described in the AVR300 documentation.
;***	(An example is shown in the 'mainloop' code).
;***
;*** NOTES
;***	The TWI routines can be called either from non-interrupt or
;***	interrupt routines, not both.
;***
;*** STATISTICS
;***	Code Size: 		81 words (maximum)
;***	Register Usage: 	4 High, 0 Low
;***	Interrupt Usage: 	None
;***	Other Usage: 		Uses two I/O pins on port D
;***	XTAL Range: 		N/A
;***
;***************************************************************************

;*** Includes

.include "..\Def\2313def.inc"			; change if an other device is used

;**** Global TWI Constants ****

.equ	SCLP	= 1			; SCL Pin number (port D)
.equ	SDAP	= 0			; SDA Pin number (port D)

.equ	b_dir	= 0			; transfer direction bit in tmpadr

.equ	TWIrd	= 1
.equ	TWIwr	= 0
.equ	IOAdr	= $40

;******************************************************************************
;***** Registers
; Register lowhalf
.def	EEdrd			= r00 ;result data byte
.def  I2CIO1		= r01
;.def         	=r02
;.def         	=r03
;.def       	=r04
;.def      	=r05
;.def      	=r06
;.def      	=r07
;.def         	=r08
;.def         	=r09
;.def         	=r10
;.def         	=r11
;.def         	=r12
;.def         	=r13
;.def         	=r14
.def  errorsave	=r15

; Register highhalf
.def	tmp			= r16		;
.def	tmpdelay		= r17		; Used as: I2C Delay loop variable
.def	tmpdata		= r18		; Used as: I2C data transfer register
.def	tmpadr		= r19		; Used as: I2C address and direction register
.def	tmpstatus	= r20		; Used as: I2C bus status register
.def	tmpcnt		= r21
.def	tmperror		= r22

;**** Interrupt Vectors ****

rjmp	MAIN			; Reset handle
rjmp  MAIN      	;Externer Int.0
rjmp  MAIN      	;Externer Int.1
rjmp  MAIN      	;Timer Int.1 event
rjmp  MAIN      	;Timer Int.1 match
rjmp  MAIN      	;Timer Int.1 overflow
rjmp  MAIN      	;Timer Int.0 overflow
rjmp  MAIN      	;UART read
rjmp  MAIN      	;UART empty
rjmp  MAIN      	;UART write
rjmp  MAIN      	;AnalogComp. Int.


main:
	;wdr
   ;ldi	tmp,0x08
   ;out  WDTCR,tmp         ;Watchdog enable 16ms Zyklen
   ldi	tmp,LOW(RAMEND)	;Initiate Stackpointer
 	out	SPL,tmp

Init:
	rcall	TWI_init		; initialize TWI interface
	; in		PORTD,tmp	;
	ldi	tmp,0b00000000	; upper nibble should be RI (for test)
								; RIs must be set to 1 before they are reread

MAINLOOP:

	sbic	PIND,2		; Check switch connected to PIN2
							; According to the position of the switch, an output
							; on the slave cicuit will be set
	ori	tmp,$01		; if switch is on, set RO1
	sbr	tmp,$F0 ; set the upper 4 bits to 1 for Input reading
	ldi	tmpadr,IOAdr+TWIwr	; Set device address and write

	rcall	I2CIO
	mov	I2CIO1,tmpdata	;save the state of IO1
	
	in		tmp,PORTD
	cbr	tmp,0b01110000
	;cpi	tmpdata,$00
	rol	tmpdata	; roll bit 6 ...
	rol	tmpdata	;... into carry
	brcs  testx1	; now set PD4 or PD5 according to the value of the Bit7 of PCF8574
	sbr	tmp,0b00010000	; Carry is not set or PCF8574 sets RI is 0
	rjmp	PDOUT
Testx1:
	sbr	tmp,0b00100000	; Carry is set  or RI is 1
	;andi	tmpdata,0b01000000
	;rol	tmpdata
	;brne LOOPFOREVER
	;sbr	tmp,0b00010000	; PD4 will be switched on, if the  I/O Line 7 of PCF8574 is 0
PDOUT:
	out	PORTD,tmp

LOOPFOREVER:
	rjmp	MAINLOOP	; Loop Forever

;***************************************************************************
;***
;*** FUNCTION
;***	TWI_hp_delay
;***	TWI_qp_delay
;***
;*** DESCRIPTION
;***	hp - half TWI clock period delay (normal: 5.0us / fast: 1.3us)
;***	qp - quarter TWI clock period delay (normal: 2.5us / fast: 0.6us)
;***
;***	SEE DOCUMENTATION !!!
;***
;*** USAGE
;***	no parameters
;***
;*** RETURN
;***	none
;***
;***************************************************************************

TWI_hp_delay:
	ldi	tmpdelay,2
TWI_hp_delay_loop:
	dec	tmpdelay
	brne	TWI_hp_delay_loop
	ret

TWI_qp_delay:
	;ldi	tmpdelay,1
TWI_qp_delay_loop:
	;dec	tmpdelay
	;brne	TWI_qp_delay_loop
	ret

;***************************************************************************
;***
;*** FUNCTION
;***	TWI_rep_start
;***
;*** DESCRIPTION
;***	Assert repeated start condition and sends slave address.
;***
;*** USAGE
;***	tmpadr - Contains the slave address and transfer direction.
;***
;*** RETURN
;***	Carry flag - Cleared if a slave responds to the address.
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_start.
;***
;***************************************************************************

TWI_rep_start:
	sbi	DDRD,SCLP		; force SCL low
	cbi	DDRD,SDAP		; release SDA
	rcall	TWI_hp_delay	; half period delay
	cbi	DDRD,SCLP		; release SCL
	rcall	TWI_qp_delay	; quarter period delay


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_start
;***
;*** DESCRIPTION
;***	Generates start condition and sends slave address.
;***
;*** USAGE
;***	tmpadr - Contains the slave address and transfer direction.
;***
;*** RETURN
;***	Carry flag - Cleared if a slave responds to the address.
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_write.
;***
;***************************************************************************

TWI_start:
	mov	tmpdata,tmpadr	; copy address to transmitt register
	sbi	DDRD,SDAP		; force SDA low
	rcall	TWI_qp_delay	; quarter period delay


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_write
;***
;*** DESCRIPTION
;***	Writes data (one byte) to the TWI bus. Also used for sending
;***	the address.
;***
;*** USAGE
;***	tmpdata - Contains data to be transmitted.
;***
;*** RETURN
;***	Carry flag - Set if the slave respond transfer.
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_get_ack.
;***
;***************************************************************************

TWI_write:
	sec						; set carry flag
	rol	tmpdata			; shift in carry and out bit one
	rjmp	TWI_write_first
TWI_write_bit:
	lsl	tmpdata			; if transmit register empty
TWI_write_first:
	breq	TWI_get_ack		; goto get acknowledge
	sbi	DDRD,SCLP		; force SCL low

	brcc	TWI_write_low	; if bit high
	nop						; (equalize number of cycles)
	cbi	DDRD,SDAP		; release SDA
	rjmp	TWI_write_high
TWI_write_low:				; else
	sbi	DDRD,SDAP		; force SDA low
	rjmp	TWI_write_high	; (equalize number of cycles)
TWI_write_high:
	rcall	TWI_hp_delay	; half period delay
	cbi	DDRD,SCLP		; release SCL
	rcall	TWI_hp_delay	; half period delay

	rjmp	TWI_write_bit

;***************************************************************************
;***
;*** FUNCTION
;***	TWI_get_ack
;***
;*** DESCRIPTION
;***	Get slave acknowledge response.
;***
;*** USAGE
;***	(used only by TWI_write in this version)
;***
;*** RETURN
;***	Carry flag - Cleared if a slave responds to a request.
;***
;***************************************************************************

TWI_get_ack:
	sbi	DDRD,SCLP		; force SCL low
	cbi	DDRD,SDAP		; release SDA
	rcall	TWI_hp_delay		; half period delay
	cbi	DDRD,SCLP		; release SCL

TWI_get_ack_wait:
	;sbis	PIND,SCLP		; wait SCL high
	sbic	PIND,SCLP		; wait SCL high
	rjmp 	TWI_ACK_OK		;(In case wait states are inserted)
	dec 	tmpcnt
	breq 	error
	rjmp	TWI_get_ack_wait
TWI_ACK_OK:
	clc						; clear carry flag
	sbic	PIND,SDAP		; if SDA is high
	sec						; set carry flag
	rcall	TWI_hp_delay	; half period delay
TWI_ACK_END:
	ret

error:
	sbr	tmperror,0b00000001 	; Bit 1 is a Signal for an I2C-Bus error
	rcall ERRORHANDLING
	pop	tmp						; get return adress from stack
	rjmp  TWI_ACK_END


ERRORHANDLING:
	;cpi	tmperror,0
	;breq	I2CWRITE_END
	sbi	PORTD,3
	or		errorsave,tmperror
	clr	tmperror
	ret


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_do_transfer
;***
;*** DESCRIPTION
;***	Executes a transfer on bus. This is only a combination of TWI_read
;***	and TWI_write for convenience.
;***
;*** USAGE
;***	tmpadr - Must have the same direction as when TWI_start was called.
;***	see TWI_read and TWI_write for more information.
;***
;*** RETURN
;***	(depends on type of transfer, read or write)
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_read.
;***
;***************************************************************************

TWI_do_transfer:
	sbrs	tmpadr,b_dir	; if dir = write
	rjmp	TWI_write		;	goto write data


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_read
;***
;*** DESCRIPTION
;***	Reads data (one byte) from the TWI bus.
;***
;*** USAGE
;***	Carry flag - 	If set no acknowledge is given to the slave
;***			indicating last read operation before a STOP.
;***			If cleared acknowledge is given to the slave
;***			indicating more data.
;***
;*** RETURN
;***	tmpdata - Contains received data.
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_put_ack.
;***
;***************************************************************************

TWI_read:
	rol	tmpstatus		; store acknowledge
					; (used by TWI_put_ack)
	ldi	tmpdata,0x01	; data = 0x01
TWI_read_bit:			; do
	sbi	DDRD,SCLP	; force SCL low
	rcall	TWI_hp_delay	; half period delay

	cbi	DDRD,SCLP	; release SCL
	rcall TWI_hp_delay	; half period delay

	clc				; clear carry flag
	sbic	PIND,SDAP	; if SDA is high
	sec				; set carry flag

	rol	tmpdata		; store data bit
	brcc	TWI_read_bit	; while receive register not full


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_put_ack
;***
;*** DESCRIPTION
;***	Put acknowledge.
;***
;*** USAGE
;***	(used only by TWI_read in this version)
;***
;*** RETURN
;***	none
;***
;***************************************************************************

TWI_put_ack:
	sbi	DDRD,SCLP			; force SCL low
	ror	tmpstatus				; get status bit
	brcc	TWI_put_ack_low	; if bit low goto assert low
	cbi	DDRD,SDAP			; release SDA
	rjmp	TWI_put_ack_high
TWI_put_ack_low:				; else
	sbi	DDRD,SDAP			; force SDA low
TWI_put_ack_high:

	rcall TWI_hp_delay		; half period delay
	cbi	 DDRD,SCLP			; release SCL
TWI_put_ack_wait:
	sbis	 PIND,SCLP			; wait SCL high
	rjmp	 TWI_put_ack_wait
	rcall TWI_hp_delay		; half period delay
	ret


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_stop
;***
;*** DESCRIPTION
;***	Assert stop condition.
;***
;*** USAGE
;***	No parameters.
;***
;*** RETURN
;***	None.
;***
;***************************************************************************

TWI_stop:
	sbi	DDRD,SCLP		; force SCL low
	sbi	DDRD,SDAP		; force SDA low
	rcall	TWI_hp_delay		; half period delay
	cbi	DDRD,SCLP		; release SCL
	rcall	TWI_qp_delay		; quarter period delay
	cbi	DDRD,SDAP		; release SDA
	rcall	TWI_hp_delay		; half period delay
	ret


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_init
;***
;*** DESCRIPTION
;***	Initialization of the TWI bus interface.
;***
;*** USAGE
;***	Call this function once to initialize the TWI bus. No parameters
;***	are required.
;***
;*** RETURN
;***	None
;***
;*** NOTE
;***	PORTD and DDRD pins not used by the TWI bus interface will be
;***	set to Hi-Z (!).
;***
;*** COMMENT
;***	This function can be combined with other PORTD initializations.
;***
;***************************************************************************
;***     Portbit-Configuration	***

;***	PORTD bits
;***		1xxxxxx	   	PD6(Pin11)
;***		x1xxxxx	   	PD5(Pin09)
;***  	xx1xxxx	Set LED if signal from RI7 PD4(Pin08)
;***  	xxx1xxx	Error Connecting I2C  		PD3(Pin07)
;***  	xxxx1xx	Switch on Remote Output   	PD2(Pin06)
;***  	xxxxx1x 	I2C SCL  						PD1(Pin03)
;***  	xxxxxx1	I2C SDA							PD0(Pin02)
;***	PORTB bits
;***		1xxxxxxx		PB7(Pin19)
;***		x1xxxxxx		PB6(Pin18)
;***		xx1xxxxx		PB5(Pin17)
;***     xxx1xxxx	   PB4(Pin16)
;***     xxxx1xxx	   PB3(Pin15)
;***     xxxxx1xx	   PB2(Pin14)
;***     xxxxxx1x	   PB1(Pin13)
;***		xxxxxxx1   	PB0(Pin12)


TWI_init:
	ldi	tmpstatus,0b01111000	; PD0 and PD1 -> Input, No Pullup
	out	DDRD,tmpstatus		; PD2 -> Input, switch on Pullup
	ldi	tmpstatus,0b00000100	; PD3,PD4 -> Output
	out	PORTD,tmpstatus		; set TWI pins to open colector
	clr	tmpstatus				; clear TWI status register (used
									; as a temporary register)
	ret



;***************************************************************************
;***
;*** FUNCTION
;***	I2CIO
;***
;*** DESCRIPTION
;***	Communcate with a PCF8574 Ciruit to expand I/O Lines
;***
;*** USAGE
;***	tmpadr - Adress of circuit
;***	are required.
;***
;*** RETURN
;***	None
;***
;*** NOTE
;***	PORTD and DDRD pins not used by the TWI bus interface will be
;***	set to Hi-Z (!).
;***
;*** COMMENT
;***	This function can be combined with other PORTD initializations.
;***
;***************************************************************************

I2CIO:

	ldi 	tmpcnt,$0F				; errorcounter, if no acknowledge
	ldi	tmpadr,IOAdr+TWIwr	; Set device address and write
	rcall TWI_start				; Send start condition and address
	mov	tmpdata,tmp	
	rcall	TWI_do_transfer		; Execute transfer
	ldi	tmpadr,IOAdr+TWIrd	; Set device address and read
	rcall	TWI_rep_start			; Send repeated start condition and address
	sec								; Set no acknowledge (read is followed by a 
										; stop condition)
	rcall	TWI_do_transfer		; Execute transfer (read)

	rcall	TWI_stop					; Send stop condition - releases bus
	
I2CEND:
	ret


;*** End of File